/** @module mat3 */
/**
 * Returns a 3x3 identity matrix.
 *
 * Row major memory layout:
 *
 * ```
 *  0   1   2
 *  3   4   5
 *  6   7   8
 * ```
 *
 *  Equivalent to the column major OpenGL spec:
 *
 * ```
 *   0   3   6
 *   1   4   7
 *   2   5   8
 *
 *  m00 m10 m20
 *  m01 m11 m21
 *  m02 m12 m22
 * ```
 * @returns {import("./types.js").mat3}
 */
export function create(): import("./types.js").mat3;
/**
 * Sets a matrix to the identity matrix.
 * @param {import("./types.js").mat3} a
 * @returns {import("./types.js").mat3}
 */
export function identity(a: import("./types.js").mat3): import("./types.js").mat3;
/**
 * Returns a copy of a matrix.
 * @param {import("./types.js").mat3} a
 * @returns {import("./types.js").mat3}
 */
export function copy(a: import("./types.js").mat3): import("./types.js").mat3;
/**
 * Sets a matrix from another matrix.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").mat3} b
 * @returns {import("./types.js").mat3}
 */
export function set(a: import("./types.js").mat3, b: import("./types.js").mat3): import("./types.js").mat3;
/**
 * Compares two matrices.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").mat3} b
 * @returns {boolean}
 */
export function equals(a: import("./types.js").mat3, b: import("./types.js").mat3): boolean;
/**
 * Multiplies two matrices.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").mat3} b
 * @returns {import("./types.js").mat3}
 */
export function mult(a: import("./types.js").mat3, b: import("./types.js").mat3): import("./types.js").mat3;
/**
 * Transposes a matrix.
 * @param {import("./types.js").mat3} a
 * @returns {import("./types.js").mat3}
 */
export function transpose(a: import("./types.js").mat3): import("./types.js").mat3;
/**
 * Sets matrix to a quaternion.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").quat} q
 * @returns {import("./types.js").mat3}
 */
export function fromQuat(a: import("./types.js").mat3, q: import("./types.js").quat): import("./types.js").mat3;
/**
 * Sets a 3x3 matrix from a 2x3 matrix.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").mat2x3} b
 * @returns {import("./types.js").mat3}
 */
export function fromMat2x3(a: import("./types.js").mat3, b: import("./types.js").mat2x3): import("./types.js").mat3;
/**
 * Sets a 3x3 matrix to a 4x4 matrix.
 * @param {import("./types.js").mat3} a
 * @param {import("./types.js").mat4} b
 * @returns {import("./types.js").mat3}
 */
export function fromMat4(a: import("./types.js").mat3, b: import("./types.js").mat4): import("./types.js").mat3;
