/**
 * Returns a 4x4 identity matrix.
 *
 * Row major memory layout:
 *
 * ```
 *  0   1   2   3
 *  4   5   6   7
 *  8   9  10  11
 * 12  13  14  15
 * ```
 *
 * Equivalent to the column major OpenGL spec:
 *
 * ```
 *  0   4   8  12
 *  1   5   9  13
 *  2   6  10  14
 *  3   7  11  15
 *
 *  m00 m10 m20 m30
 *  m01 m11 m21 m31
 *  m02 m12 m22 m32
 *  m03 m13 m23 m33
 * ```
 * @returns {import("./types.js").mat4}
 */
export function create(): import("./types.js").mat4;
/**
 * Sets a matrix to the identity matrix.
 * @param {import("./types.js").mat4} a
 * @returns {import("./types.js").mat4}
 */
export function identity(a: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Returns a copy of a matrix.
 * @param {import("./types.js").mat4} a
 * @returns {import("./types.js").mat4}
 */
export function copy(a: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Sets a matrix from another matrix.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").mat4} b
 * @returns {import("./types.js").mat4}
 */
export function set(a: import("./types.js").mat4, b: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Compares two matrices.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").mat4} b
 * @returns {boolean}
 */
export function equals(a: import("./types.js").mat4, b: import("./types.js").mat4): boolean;
/**
 * Multiplies two matrices.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").mat4} b
 * @returns {import("./types.js").mat4}
 */
export function mult(a: import("./types.js").mat4, b: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Inverts a matrix.
 * @param {import("./types.js").mat4} a
 * @returns {import("./types.js").mat4}
 */
export function invert(a: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Transposes a matrix.
 * @param {import("./types.js").mat4} a
 * @returns {import("./types.js").mat4}
 */
export function transpose(a: import("./types.js").mat4): import("./types.js").mat4;
/**
 * Translates a matrix by a vector.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").vec3} v
 * @returns {import("./types.js").mat4}
 */
export function translate(a: import("./types.js").mat4, [x, y, z]: import("./types.js").vec3): import("./types.js").mat4;
/**
 * Rotates a matrix by an angle at an axis.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").Radians} r
 * @param {import("./types.js").vec3} v
 * @returns {import("./types.js").mat4}
 */
export function rotate(a: import("./types.js").mat4, r: import("./types.js").Radians, [x, y, z]: import("./types.js").vec3): import("./types.js").mat4;
/**
 * Scales a matrix by a vector.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").vec3} v
 * @returns {import("./types.js").mat4}
 */
export function scale(a: import("./types.js").mat4, [x, y, z]: import("./types.js").vec3): import("./types.js").mat4;
/**
 * Sets a matrix to a quaternion.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").quat} q
 * @returns {import("./types.js").mat4}
 */
export function fromQuat(a: import("./types.js").mat4, q: import("./types.js").quat): import("./types.js").mat4;
/**
 * Sets a matrix to the TRS matrix.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").vec3} v
 * @param {import("./types.js").quat} q
 * @param {import("./types.js").vec3} s
 * @returns {import("./types.js").mat4}
 */
export function fromTranslationRotationScale(a: import("./types.js").mat4, v: import("./types.js").vec3, q: import("./types.js").quat, s: import("./types.js").vec3): import("./types.js").mat4;
/**
 * Sets a 4x4 matrix to a 3x3 matrix.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").mat3} b
 * @returns {import("./types.js").mat4}
 */
export function fromMat3(a: import("./types.js").mat4, b: import("./types.js").mat3): import("./types.js").mat4;
/**
 * Creates a frustum matrix.
 * @param {import("./types.js").mat4} a
 * @param {number} left
 * @param {number} right
 * @param {number} bottom
 * @param {number} top
 * @param {number} near
 * @param {number} far
 * @returns {import("./types.js").mat4}
 */
export function frustum(a: import("./types.js").mat4, left: number, right: number, bottom: number, top: number, near: number, far: number): import("./types.js").mat4;
/**
 * Creates a perspective matrix.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").Radians} fovy
 * @param {number} aspectRatio
 * @param {number} near
 * @param {number} far
 * @returns {import("./types.js").mat4}
 */
export function perspective(a: import("./types.js").mat4, fovy: import("./types.js").Radians, aspectRatio: number, near: number, far: number): import("./types.js").mat4;
/**
 * Creates an orthographic matrix.
 * @param {import("./types.js").mat4} a
 * @param {number} left
 * @param {number} right
 * @param {number} bottom
 * @param {number} top
 * @param {number} near
 * @param {number} far
 * @returns {import("./types.js").mat4}
 */
export function ortho(a: import("./types.js").mat4, left: number, right: number, bottom: number, top: number, near: number, far: number): import("./types.js").mat4;
/**
 * Calculates a lookAt matrix from position, target and up vectors.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").vec3} from
 * @param {import("./types.js").vec3} to
 * @param {import("./types.js").vec3} [up=Y_UP]
 * @returns {import("./types.js").mat4}
 */
export function lookAt(a: import("./types.js").mat4, [eyex, eyey, eyez]: import("./types.js").vec3, [targetx, targety, targetz]: import("./types.js").vec3, [upx, upy, upz]?: import("./types.js").vec3): import("./types.js").mat4;
/**
 * Sets a matrix from a vector to another.
 * @param {import("./types.js").mat4} a
 * @param {import("./types.js").vec3} from
 * @param {import("./types.js").vec3} to
 * @param {import("./types.js").vec3} [up=Y_UP]
 * @returns {import("./types.js").mat4}
 */
export function targetTo(a: import("./types.js").mat4, [eyex, eyey, eyez]: import("./types.js").vec3, [targetx, targety, targetz]: import("./types.js").vec3, [upx, upy, upz]?: import("./types.js").vec3): import("./types.js").mat4;
