/** @module vec3 */
/**
 * Returns a new vec3 at 0, 0, 0.
 * @returns {import("./types.js").vec3}
 */
export function create(): import("./types.js").vec3;
/**
 * Returns a copy of a vector.
 * @param {import("./types.js").vec3} a
 * @returns {import("./types.js").vec3}
 */
export function copy(a: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Sets a vector to another vector.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {import("./types.js").vec3}
 */
export function set(a: import("./types.js").vec3, b: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Compares two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {boolean}
 */
export function equals(a: import("./types.js").vec3, b: import("./types.js").vec3): boolean;
/**
 * Adds a vector to another.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {import("./types.js").vec3}
 */
export function add(a: import("./types.js").vec3, b: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Subtracts a vector from another.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {import("./types.js").vec3}
 */
export function sub(a: import("./types.js").vec3, b: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Scales a vector by a number.
 * @param {import("./types.js").vec3} a
 * @param {number} s
 * @returns {import("./types.js").vec3}
 */
export function scale(a: import("./types.js").vec3, s: number): import("./types.js").vec3;
/**
 * Adds two vectors after scaling the second one.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @param {number} s
 * @returns {import("./types.js").vec3}
 */
export function addScaled(a: import("./types.js").vec3, b: import("./types.js").vec3, s: number): import("./types.js").vec3;
/**
 * Multiplies a vector by a matrix.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").mat4} m
 * @returns {import("./types.js").vec3}
 */
export function multMat4(a: import("./types.js").vec3, m: import("./types.js").mat4): import("./types.js").vec3;
/**
 * Multiplies a vector by a quaternion.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").quat} q
 * @returns {import("./types.js").vec3}
 */
export function multQuat(a: import("./types.js").vec3, q: import("./types.js").quat): import("./types.js").vec3;
/**
 * Calculates the dot product of two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {number}
 */
export function dot(a: import("./types.js").vec3, b: import("./types.js").vec3): number;
/**
 * Calculates the cross product of two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {import("./types.js").vec3}
 */
export function cross(a: import("./types.js").vec3, b: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Calculates the length of a vector.
 * @param {import("./types.js").vec3} a
 * @returns {number}
 */
export function length(a: import("./types.js").vec3): number;
/**
 * Calculates the squared length of a vector.
 * @param {import("./types.js").vec3} a
 * @returns {number}
 */
export function lengthSq(a: import("./types.js").vec3): number;
/**
 * Normalises a vector.
 * @param {import("./types.js").vec3} a
 * @returns {import("./types.js").vec3}
 */
export function normalize(a: import("./types.js").vec3): import("./types.js").vec3;
/**
 * Calculates the distance between two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {number}
 */
export function distance(a: import("./types.js").vec3, b: import("./types.js").vec3): number;
/**
 * Calculates the squared distance between two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @returns {number}
 */
export function distanceSq(a: import("./types.js").vec3, b: import("./types.js").vec3): number;
/**
 * Limits a vector to a length.
 * @param {import("./types.js").vec3} a
 * @param {number} len
 * @returns {import("./types.js").vec3}
 */
export function limit(a: import("./types.js").vec3, len: number): import("./types.js").vec3;
/**
 * Linearly interpolates between two vectors.
 * @param {import("./types.js").vec3} a
 * @param {import("./types.js").vec3} b
 * @param {number} t
 * @returns {import("./types.js").vec3}
 */
export function lerp(a: import("./types.js").vec3, b: import("./types.js").vec3, t: number): import("./types.js").vec3;
/**
 * Prints a vector to a string.
 * @param {import("./types.js").vec3} a
 * @param {number} [precision=4]
 * @returns {string}
 */
export function toString(a: import("./types.js").vec3, precision?: number): string;
